@@ -256,8 +256,18 @@ def setup(app):
                         objname='runtime system command-line option',
                         parse_node=parse_flag,
                         indextemplate='pair: %s; RTS option',
+                        doc_field_types=[
+                            Field('since', label='Introduced in GHC version', names=['since'])
+                        ])
+
+    app.add_object_type('event-type', 'event-type',
+                        objname='event log event type',
+                        indextemplate='pair: %s; eventlog event type',
                         doc_field_types=[
                             Field('since', label='Introduced in GHC version', names=['since']),
+                            Field('tag', label='Event type ID', names=['tag']),
+                            Field('length', label='Record length', names=['length']),
+                            TypedField('fields', label='Fields', names='field', typenames=('fieldtype', 'type'))
                         ])
 
     app.add_object_type('pragma', 'pragma',
