@@ -8,12 +8,11 @@
  * Do not #include this file directly: #include "Rts.h" instead.
  *
  * To understand the structure of the RTS headers, see the wiki:
- *   http://ghc.haskell.org/trac/ghc/wiki/Commentary/SourceTree/Includes
+ *   https://gitlab.haskell.org/ghc/ghc/wikis/commentary/source-tree/includes
  *
  * ---------------------------------------------------------------------------*/
 
-#ifndef MACHREGS_H
-#define MACHREGS_H
+#pragma once
 
 /* This file is #included into Haskell code in the compiler: #defines
  * only in here please.
@@ -82,16 +81,16 @@
    Leaving SpLim out of the picture.
    -------------------------------------------------------------------------- */
 
-#if MACHREGS_i386
+#if defined(MACHREGS_i386)
 
 #define REG(x) __asm__("%" #x)
 
-#ifndef not_doing_dynamic_linking
+#if !defined(not_doing_dynamic_linking)
 #define REG_Base    ebx
 #endif
 #define REG_Sp      ebp
 
-#ifndef STOLEN_X86_REGS
+#if !defined(STOLEN_X86_REGS)
 #define STOLEN_X86_REGS 4
 #endif
 
@@ -102,6 +101,7 @@
 #if STOLEN_X86_REGS >= 4
 # define REG_Hp     edi
 #endif
+#define REG_MachSp  esp
 
 #define REG_XMM1    xmm0
 #define REG_XMM2    xmm1
@@ -155,7 +155,7 @@
 
   --------------------------------------------------------------------------- */
 
-#elif MACHREGS_x86_64
+#elif defined(MACHREGS_x86_64)
 
 #define REG(x) __asm__("%" #x)
 
@@ -169,6 +169,13 @@
 #define REG_R5    r8
 #define REG_R6    r9
 #define REG_SpLim r15
+#define REG_MachSp  rsp
+
+/*
+Map both Fn and Dn to register xmmn so that we can pass a function any
+combination of up to six Float# or Double# arguments without touching
+the stack. See Note [Overlapping global registers] for implications.
+*/
 
 #define REG_F1    xmm1
 #define REG_F2    xmm2
@@ -272,8 +279,6 @@
    1            SP              (callee-save, non-volatile)
    2            AIX, powerpc64-linux:
                     RTOC        (a strange special case)
-                darwin:
-                                (caller-save, volatile)
                 powerpc32-linux:
                                 reserved for use by system
 
@@ -295,7 +300,7 @@
    We can do the Whole Business with callee-save registers only!
    -------------------------------------------------------------------------- */
 
-#elif MACHREGS_powerpc
+#elif defined(MACHREGS_powerpc)
 
 #define REG(x) __asm__(#x)
 
@@ -307,36 +312,31 @@
 #define REG_R6          r19
 #define REG_R7          r20
 #define REG_R8          r21
-
-#if MACHREGS_darwin
-
-#define REG_F1          f14
-#define REG_F2          f15
-#define REG_F3          f16
-#define REG_F4          f17
-
-#define REG_D1          f18
-#define REG_D2          f19
-
-#else
+#define REG_R9          r22
+#define REG_R10         r23
 
 #define REG_F1          fr14
 #define REG_F2          fr15
 #define REG_F3          fr16
 #define REG_F4          fr17
+#define REG_F5          fr18
+#define REG_F6          fr19
 
-#define REG_D1          fr18
-#define REG_D2          fr19
-
-#endif
-
-#define REG_Sp          r22
-#define REG_SpLim       r24
-
-#define REG_Hp          r25
+#define REG_D1          fr20
+#define REG_D2          fr21
+#define REG_D3          fr22
+#define REG_D4          fr23
+#define REG_D5          fr24
+#define REG_D6          fr25
 
+#define REG_Sp          r24
+#define REG_SpLim       r25
+#define REG_Hp          r26
 #define REG_Base        r27
 
+#define MAX_REAL_FLOAT_REG   6
+#define MAX_REAL_DOUBLE_REG  6
+
 /* -----------------------------------------------------------------------------
    The Sun SPARC register mapping
 
@@ -428,7 +428,7 @@
 
    -------------------------------------------------------------------------- */
 
-#elif MACHREGS_sparc
+#elif defined(MACHREGS_sparc)
 
 #define REG(x) __asm__("%" #x)
 
@@ -477,14 +477,15 @@
    Here we consider ARM mode (i.e. 32bit isns)
    and also CPU with full VFPv3 implementation
 
-   ARM registers (see Chapter 5.1 in ARM IHI 0042D)
+   ARM registers (see Chapter 5.1 in ARM IHI 0042D and
+   Section 9.2.2 in ARM Software Development Toolkit Reference Guide)
 
    r15  PC         The Program Counter.
    r14  LR         The Link Register.
    r13  SP         The Stack Pointer.
    r12  IP         The Intra-Procedure-call scratch register.
-   r11  v8         Variable-register 8.
-   r10  v7         Variable-register 7.
+   r11  v8/fp      Variable-register 8.
+   r10  v7/sl      Variable-register 7.
    r9   v6/SB/TR   Platform register. The meaning of this register is
                    defined by the platform standard.
    r8   v5         Variable-register 5.
@@ -500,13 +501,13 @@
    VFPv2/VFPv3/NEON registers
    s0-s15/d0-d7/q0-q3    Argument / result/ scratch registers
    s16-s31/d8-d15/q4-q7  callee-saved registers (must be preserved across
-                         subrutine calls)
+                         subroutine calls)
 
    VFPv3/NEON registers (added to the VFPv2 registers set)
    d16-d31/q8-q15        Argument / result/ scratch registers
    ----------------------------------------------------------------------------- */
 
-#elif MACHREGS_arm
+#elif defined(MACHREGS_arm)
 
 #define REG(x) __asm__(#x)
 
@@ -557,13 +558,13 @@
    FPU/SIMD registers
 
    s/d/q/v0-v7    Argument / result/ scratch registers
-   s/d/q/v8-v15   callee-saved registers (must be preserved across subrutine calls,
+   s/d/q/v8-v15   callee-saved registers (must be preserved across subroutine calls,
                   but only bottom 64-bit value needs to be preserved)
    s/d/q/v16-v31  temporary registers
 
    ----------------------------------------------------------------------------- */
 
-#elif MACHREGS_aarch64
+#elif defined(MACHREGS_aarch64)
 
 #define REG(x) __asm__(#x)
 
@@ -588,6 +589,78 @@
 #define REG_D3          d14
 #define REG_D4          d15
 
+/* -----------------------------------------------------------------------------
+   The s390x register mapping
+
+   Register    | Role(s)                                 | Call effect
+   ------------+-------------------------------------+-----------------
+   r0,r1       | -                                       | caller-saved
+   r2          | Argument / return value                 | caller-saved
+   r3,r4,r5    | Arguments                               | caller-saved
+   r6          | Argument                                | callee-saved
+   r7...r11    | -                                       | callee-saved
+   r12         | (Commonly used as GOT pointer)          | callee-saved
+   r13         | (Commonly used as literal pool pointer) | callee-saved
+   r14         | Return address                          | caller-saved
+   r15         | Stack pointer                           | callee-saved
+   f0          | Argument / return value                 | caller-saved
+   f2,f4,f6    | Arguments                               | caller-saved
+   f1,f3,f5,f7 | -                                       | caller-saved
+   f8...f15    | -                                       | callee-saved
+   v0...v31    | -                                       | caller-saved
+
+   Each general purpose register r0 through r15 as well as each floating-point
+   register f0 through f15 is 64 bits wide. Each vector register v0 through v31
+   is 128 bits wide.
+
+   Note, the vector registers v0 through v15 overlap with the floating-point
+   registers f0 through f15.
+
+   -------------------------------------------------------------------------- */
+
+#elif defined(MACHREGS_s390x)
+
+#define REG(x) __asm__("%" #x)
+
+#define REG_Base        r7
+#define REG_Sp          r8
+#define REG_Hp          r10
+#define REG_R1          r11
+#define REG_R2          r12
+#define REG_R3          r13
+#define REG_R4          r6
+#define REG_R5          r2
+#define REG_R6          r3
+#define REG_R7          r4
+#define REG_R8          r5
+#define REG_SpLim       r9
+#define REG_MachSp      r15
+
+#define REG_F1          f8
+#define REG_F2          f9
+#define REG_F3          f10
+#define REG_F4          f11
+#define REG_F5          f0
+#define REG_F6          f1
+
+#define REG_D1          f12
+#define REG_D2          f13
+#define REG_D3          f14
+#define REG_D4          f15
+#define REG_D5          f2
+#define REG_D6          f3
+
+#define CALLER_SAVES_R5
+#define CALLER_SAVES_R6
+#define CALLER_SAVES_R7
+#define CALLER_SAVES_R8
+
+#define CALLER_SAVES_F5
+#define CALLER_SAVES_F6
+
+#define CALLER_SAVES_D5
+#define CALLER_SAVES_D6
+
 #else
 
 #error Cannot find platform to give register info for
@@ -615,7 +688,7 @@
  * communicate with PrimOps and RTS functions.
  */
 
-#ifndef MAX_REAL_VANILLA_REG
+#if !defined(MAX_REAL_VANILLA_REG)
 #  if   defined(REG_R10)
 #  define MAX_REAL_VANILLA_REG 10
 #  elif   defined(REG_R9)
@@ -641,7 +714,7 @@
 #  endif
 #endif
 
-#ifndef MAX_REAL_FLOAT_REG
+#if !defined(MAX_REAL_FLOAT_REG)
 #  if   defined(REG_F4)
 #  define MAX_REAL_FLOAT_REG 4
 #  elif defined(REG_F3)
@@ -655,7 +728,7 @@
 #  endif
 #endif
 
-#ifndef MAX_REAL_DOUBLE_REG
+#if !defined(MAX_REAL_DOUBLE_REG)
 #  if   defined(REG_D2)
 #  define MAX_REAL_DOUBLE_REG 2
 #  elif defined(REG_D1)
@@ -665,7 +738,7 @@
 #  endif
 #endif
 
-#ifndef MAX_REAL_LONG_REG
+#if !defined(MAX_REAL_LONG_REG)
 #  if   defined(REG_L1)
 #  define MAX_REAL_LONG_REG 1
 #  else
@@ -673,7 +746,7 @@
 #  endif
 #endif
 
-#ifndef MAX_REAL_XMM_REG
+#if !defined(MAX_REAL_XMM_REG)
 #  if   defined(REG_XMM6)
 #  define MAX_REAL_XMM_REG 6
 #  elif defined(REG_XMM5)
@@ -699,5 +772,3 @@
 #else
 #undef NO_ARG_REGS
 #endif
-
-#endif /* MACHREGS_H */
