@@ -58,6 +58,7 @@ void      stat_endHeapCensus(void);
 void      stat_startExit(void);
 void      stat_endExit(void);
 
+void      stat_exitReport(void);
 void      stat_exit(void);
 void      stat_workerStop(void);
 
@@ -65,7 +66,6 @@ void      initStats0(void);
 void      initStats1(void);
 void      resetChildProcessStats(void);
 
-double    mut_user_time_until(Time t);
 double    mut_user_time(void);
 
 void      statDescribeGens( void );
