@@ -30,6 +30,7 @@
 Task *all_tasks = NULL;
 
 // current number of bound tasks + total number of worker tasks.
+// Locks required: all_tasks_mutex.
 uint32_t taskCount;
 uint32_t workerCount;
 uint32_t currentWorkerCount;
@@ -199,6 +200,7 @@ freeTask (Task *task)
     stgFree(task);
 }
 
+/* Must take all_tasks_mutex */
 static Task*
 newTask (bool worker)
 {
@@ -219,6 +221,7 @@ newTask (bool worker)
 #if defined(THREADED_RTS)
     initCondition(&task->cond);
     initMutex(&task->lock);
+    task->id = 0;
     task->wakeup = false;
     task->node = 0;
 #endif
@@ -235,6 +238,8 @@ newTask (bool worker)
     all_tasks = task;
 
     taskCount++;
+    debugTrace(DEBUG_sched, "new task (taskCount: %d)", taskCount);
+
     if (worker) {
         workerCount++;
         currentWorkerCount++;
@@ -309,8 +314,6 @@ newBoundTask (void)
     task->stopped = false;
 
     newInCall(task);
-
-    debugTrace(DEBUG_sched, "new task (taskCount: %d)", taskCount);
     return task;
 }
 
@@ -415,7 +418,7 @@ workerTaskStop (Task *task)
 
 #if defined(THREADED_RTS)
 
-static void OSThreadProcAttr
+static void* OSThreadProcAttr
 workerStart(Task *task)
 {
     Capability *cap;
@@ -441,8 +444,11 @@ workerStart(Task *task)
     traceTaskCreate(task, cap);
 
     scheduleWorker(cap,task);
+
+    return NULL;
 }
 
+/* N.B. must take all_tasks_mutex */
 void
 startWorkerTask (Capability *cap)
 {
@@ -468,7 +474,7 @@ startWorkerTask (Capability *cap)
   // else get in, because the new worker Task has nowhere to go to
   // sleep so that it could be woken up again.
   ASSERT_LOCK_HELD(&cap->lock);
-  cap->running_task = task;
+  RELAXED_STORE(&cap->running_task, task);
 
   // Set the name of the worker thread to the original process name followed by
   // ":w", but only if we're on Linux where the program_invocation_short_name
